package java.net;

import sun.misc.JavaNetAccess;
import java.security.Permission;
import sun.net.www.ParseUtil;
import java.io.File;
import java.io.FilePermission;
import java.security.PermissionCollection;
import java.util.NoSuchElementException;
import java.util.Enumeration;
import java.security.PrivilegedAction;
import java.util.jar.Attributes;
import sun.misc.SharedSecrets;
import java.nio.ByteBuffer;
import sun.misc.PerfCounter;
import java.security.CodeSource;
import java.util.jar.Manifest;
import java.security.PrivilegedActionException;
import sun.misc.Resource;
import java.security.PrivilegedExceptionAction;
import java.util.Iterator;
import java.util.List;
import java.util.jar.JarFile;
import java.io.IOException;
import sun.net.www.protocol.file.FileURLConnection;
import java.io.InputStream;
import java.security.AccessController;
import java.util.WeakHashMap;
import java.security.AccessControlContext;
import sun.misc.URLClassPath;
import java.io.Closeable;
import java.security.SecureClassLoader;

public class URLClassLoader extends SecureClassLoader implements Closeable
{
    private final URLClassPath ucp;
    private final AccessControlContext acc;
    private WeakHashMap<Closeable, Void> closeables;
    
    public URLClassLoader(URL[] array, ClassLoader parent) {
        super(parent);
        this.closeables = new WeakHashMap();
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager != null) {
            securityManager.checkCreateClassLoader();
        }
        this.acc = AccessController.getContext();
        this.ucp = new URLClassPath(array, this.acc);
    }
    
    URLClassLoader(URL[] array, ClassLoader parent, AccessControlContext acc) {
        super(parent);
        this.closeables = new WeakHashMap();
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager != null) {
            securityManager.checkCreateClassLoader();
        }
        this.acc = acc;
        this.ucp = new URLClassPath(array, acc);
    }
    
    public URLClassLoader(URL[] array) {
        this.closeables = new WeakHashMap();
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager != null) {
            securityManager.checkCreateClassLoader();
        }
        this.acc = AccessController.getContext();
        this.ucp = new URLClassPath(array, this.acc);
    }
    
    URLClassLoader(URL[] array, AccessControlContext acc) {
        this.closeables = new WeakHashMap();
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager != null) {
            securityManager.checkCreateClassLoader();
        }
        this.acc = acc;
        this.ucp = new URLClassPath(array, acc);
    }
    
    public URLClassLoader(URL[] array, ClassLoader parent, URLStreamHandlerFactory urlStreamHandlerFactory) {
        super(parent);
        this.closeables = new WeakHashMap();
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager != null) {
            securityManager.checkCreateClassLoader();
        }
        this.acc = AccessController.getContext();
        this.ucp = new URLClassPath(array, urlStreamHandlerFactory, this.acc);
    }
    
    @Override
    public InputStream getResourceAsStream(String name) {
        URL resource = this.getResource(name);
        try {
            if (resource == null) {
                return null;
            }
            URLConnection openConnection = resource.openConnection();
            InputStream inputStream = openConnection.getInputStream();
            if ((openConnection instanceof JarURLConnection)) {
                JarFile jarFile = ((JarURLConnection)openConnection).getJarFile();
                synchronized (this.closeables) {
                    if (!this.closeables.containsKey(jarFile)) {
                        this.closeables.put(jarFile, null);
                    }
                }
            }
            else if ((openConnection instanceof FileURLConnection)) {
                synchronized (this.closeables) {
                    this.closeables.put(inputStream, null);
                }
            }
            return inputStream;
        }
        catch (IOException ex) {
            return null;
        }
    }
    
    @Override
    public void close() throws IOException {
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager != null) {
            securityManager.checkPermission(new RuntimePermission("closeClassLoader"));
        }
        List<IOException> closeLoaders = this.ucp.closeLoaders();
        synchronized (this.closeables) {
            for (Closeable closeable : this.closeables.keySet()) {
                try {
                    closeable.close();
                }
                catch (IOException ex) {
                    closeLoaders.add(ex);
                }
            }
            this.closeables.clear();
        }
        if (closeLoaders.isEmpty()) {
            return;
        }
        IOException ex2 = (IOException)closeLoaders.remove(0);
        Iterator<Object> iterator2 = closeLoaders.iterator();
        while (iterator2.hasNext()) {
            ex2.addSuppressed((IOException)iterator2.next());
        }
        throw ex2;
    }
    
    protected void addURL(URL url) {
        this.ucp.addURL(url);
    }
    
    public URL[] getURLs() {
        return this.ucp.getURLs();
    }
    
    @Override
    protected Class<?> findClass(final String s) throws ClassNotFoundException {
        Class<Object> clazz;
        try {
            clazz = (Class)AccessController.doPrivileged(new PrivilegedExceptionAction<Class<?>>() {
                @Override
                public Class<?> run() throws ClassNotFoundException {
                    Resource resource = URLClassLoader.this.ucp.getResource(s.replace('.', '/').concat(".class"), false);
                    if (resource != null) {
                        try {
                            return URLClassLoader.access$100(URLClassLoader.this, s, resource);
                        }
                        catch (IOException ex) {
                            throw new ClassNotFoundException(s, ex);
                        }
                    }
                    else {
                        return null;
                    }
                }
            }, this.acc);
        }
        catch (PrivilegedActionException ex) {
            throw (ClassNotFoundException)ex.getException();
        }
        if (clazz == null) {
            throw new ClassNotFoundException(s);
        }
        return clazz;
    }
    
    private Package getAndVerifyPackage(String str, Manifest manifest, URL url) {
        Package package1 = this.getPackage(str);
        if (package1 != null) {
            if (package1.isSealed()) {
                if (!package1.isSealed(url)) {
                    throw new SecurityException("sealing violation: package " + str + " is sealed");
                }
            }
            else if (manifest != null && this.isSealed(str, manifest)) {
                throw new SecurityException("sealing violation: can't seal package " + str + ": already loaded");
            }
        }
        return package1;
    }
    
    private void definePackageInternal(String s, Manifest manifest, URL url) {
        if (this.getAndVerifyPackage(s, manifest, url) == null) {
            try {
                if (manifest != null) {
                    this.definePackage(s, manifest, url);
                }
                else {
                    this.definePackage(s, null, null, null, null, null, null, null);
                }
            }
            catch (IllegalArgumentException ex) {
                if (this.getAndVerifyPackage(s, manifest, url) == null) {
                    throw new AssertionError("Cannot find package " + s);
                }
            }
        }
    }
    
    private Class<?> defineClass(String s, Resource resource) throws IOException {
        long nanoTime = System.nanoTime();
        int lastIndex = s.lastIndexOf(46);
        URL codeSourceURL = resource.getCodeSourceURL();
        if (lastIndex != -1) {
            this.definePackageInternal(s.substring(0, lastIndex), resource.getManifest(), codeSourceURL);
        }
        ByteBuffer byteBuffer = resource.getByteBuffer();
        if (byteBuffer != null) {
            CodeSource cs2 = new CodeSource(codeSourceURL, resource.getCodeSigners());
            PerfCounter.getReadClassBytesTime().addElapsedTimeFrom(nanoTime);
            return super.defineClass(s, byteBuffer, cs2);
        }
        byte[] bytes = resource.getBytes();
        CodeSource cs = new CodeSource(codeSourceURL, resource.getCodeSigners());
        PerfCounter.getReadClassBytesTime().addElapsedTimeFrom(nanoTime);
        return super.defineClass(s, bytes, 0, bytes.length, cs);
    }
    
    protected Package definePackage(String name, Manifest manifest, URL url) throws IllegalArgumentException {
        String specTitle = null;
        String specVersion = null;
        String specVendor = null;
        String implTitle = null;
        String implVersion = null;
        String implVendor = null;
        String anotherString = null;
        URL sealBase = null;
        Attributes trustedAttributes = SharedSecrets.javaUtilJarAccess().getTrustedAttributes(manifest, name.replace('.', '/').concat("/"));
        if (trustedAttributes != null) {
            specTitle = trustedAttributes.getValue(Attributes.Name.SPECIFICATION_TITLE);
            specVersion = trustedAttributes.getValue(Attributes.Name.SPECIFICATION_VERSION);
            specVendor = trustedAttributes.getValue(Attributes.Name.SPECIFICATION_VENDOR);
            implTitle = trustedAttributes.getValue(Attributes.Name.IMPLEMENTATION_TITLE);
            implVersion = trustedAttributes.getValue(Attributes.Name.IMPLEMENTATION_VERSION);
            implVendor = trustedAttributes.getValue(Attributes.Name.IMPLEMENTATION_VENDOR);
            anotherString = trustedAttributes.getValue(Attributes.Name.SEALED);
        }
        Attributes mainAttributes = manifest.getMainAttributes();
        if (mainAttributes != null) {
            if (specTitle == null) {
                specTitle = mainAttributes.getValue(Attributes.Name.SPECIFICATION_TITLE);
            }
            if (specVersion == null) {
                specVersion = mainAttributes.getValue(Attributes.Name.SPECIFICATION_VERSION);
            }
            if (specVendor == null) {
                specVendor = mainAttributes.getValue(Attributes.Name.SPECIFICATION_VENDOR);
            }
            if (implTitle == null) {
                implTitle = mainAttributes.getValue(Attributes.Name.IMPLEMENTATION_TITLE);
            }
            if (implVersion == null) {
                implVersion = mainAttributes.getValue(Attributes.Name.IMPLEMENTATION_VERSION);
            }
            if (implVendor == null) {
                implVendor = mainAttributes.getValue(Attributes.Name.IMPLEMENTATION_VENDOR);
            }
            if (anotherString == null) {
                anotherString = mainAttributes.getValue(Attributes.Name.SEALED);
            }
        }
        if ("true".equalsIgnoreCase(anotherString)) {
            sealBase = url;
        }
        return this.definePackage(name, specTitle, specVersion, specVendor, implTitle, implVersion, implVendor, sealBase);
    }
    
    private boolean isSealed(String s, Manifest manifest) {
        Attributes trustedAttributes = SharedSecrets.javaUtilJarAccess().getTrustedAttributes(manifest, s.replace('.', '/').concat("/"));
        String anotherString = null;
        if (trustedAttributes != null) {
            anotherString = trustedAttributes.getValue(Attributes.Name.SEALED);
        }
        Attributes mainAttributes;
        if (anotherString == null && (mainAttributes = manifest.getMainAttributes()) != null) {
            anotherString = mainAttributes.getValue(Attributes.Name.SEALED);
        }
        return "true".equalsIgnoreCase(anotherString);
    }
    
    public URL findResource(final String s) {
        URL url = (URL)AccessController.doPrivileged(new PrivilegedAction<URL>() {
            @Override
            public URL run() {
                return URLClassLoader.this.ucp.findResource(s, true);
            }
        }, this.acc);
        return (url != null) ? this.ucp.checkURL(url) : null;
    }
    
    public Enumeration<URL> findResources(String s) throws IOException {
        return new Enumeration<URL>() {
            private URL url = null;
            
            private boolean next() {
                if (this.url != null) {
                    return true;
                }
                do {
                    URL url = (URL)AccessController.doPrivileged(new PrivilegedAction<URL>() {
                        @Override
                        public URL run() {
                            if (!e.hasMoreElements()) {
                                return null;
                            }
                            return (URL)e.nextElement();
                        }
                    }, URLClassLoader.this.acc);
                    if (url == null) {
                        break;
                    }
                    this.url = URLClassLoader.this.ucp.checkURL(url);
                } while (this.url == null);
                return this.url != null;
            }
            
            @Override
            public URL nextElement() {
                if (!this.next()) {
                    throw new NoSuchElementException();
                }
                URL url = this.url;
                this.url = null;
                return url;
            }
            
            @Override
            public boolean hasMoreElements() {
                return this.next();
            }
        };
    }
    
    @Override
    protected PermissionCollection getPermissions(CodeSource codesource) {
        PermissionCollection permissions = super.getPermissions(codesource);
        URL url = codesource.getLocation();
        URLConnection openConnection;
        final Object permission;
        try {
            openConnection = url.openConnection();
            permission = openConnection.getPermission();
        }
        catch (IOException ex) {
            permission = null;
            openConnection = null;
        }
        if ((permission instanceof FilePermission)) {
            String name = permission.getName();
            if (name.endsWith(File.separator)) {
                permission = new FilePermission(name + "-", "read");
            }
        }
        else if (permission == null && url.getProtocol().equals("file")) {
            String s = ParseUtil.decode(url.getFile().replace('/', File.separatorChar));
            if (s.endsWith(File.separator)) {
                s += "-";
            }
            permission = new FilePermission(s, "read");
        }
        else {
            if ((openConnection instanceof JarURLConnection)) {
                url = ((JarURLConnection)openConnection).getJarFileURL();
            }
            String host = url.getHost();
            if (host != null && host.length() > 0) {
                permission = new SocketPermission(host, "connect,accept");
            }
        }
        if (permission != null) {
            final SecurityManager securityManager = System.getSecurityManager();
            if (securityManager != null) {
                AccessController.doPrivileged(new PrivilegedAction<Void>() {
                    @Override
                    public Void run() throws SecurityException {
                        securityManager.checkPermission(permission);
                        return null;
                    }
                }, this.acc);
            }
            permissions.add(permission);
            return permissions;
        }
        else {
            return permissions;
        }
    }
    
    public static URLClassLoader newInstance(final URL[] array, final ClassLoader classLoader) {
        return (URLClassLoader)AccessController.doPrivileged(new PrivilegedAction<URLClassLoader>() {
            
            @Override
            public URLClassLoader run() {
                return new FactoryURLClassLoader(array, classLoader, acc);
            }
        });
    }
    
    public static URLClassLoader newInstance(final URL[] array) {
        return (URLClassLoader)AccessController.doPrivileged(new PrivilegedAction<URLClassLoader>() {
            
            @Override
            public URLClassLoader run() {
                return new FactoryURLClassLoader(array, acc);
            }
        });
    }
    
    static Class access$100(URLClassLoader urlClassLoader, String s, Resource resource) throws IOException {
        return urlClassLoader.defineClass(s, resource);
    }
    
    static {
        SharedSecrets.setJavaNetAccess(new JavaNetAccess() {
            public URLClassPath getURLClassPath(URLClassLoader urlClassLoader) {
                return urlClassLoader.ucp;
            }
            
            public String getOriginalHostName(InetAddress inetAddress) {
                return inetAddress.holder.getOriginalHostName();
            }
        });
        ClassLoader.registerAsParallelCapable();
    }
}
